
#ifndef PARSE_TYPES_H
#define PARSE_TYPES_H

#include "buffer.h" 

#include "sl_list.h" 

#include "axis-types.h" 

#include "bitset.h"

struct YYLTYPE; // pre-declare

namespace atlas
{
  namespace interpreter
  {


struct expr;

typedef Hash_table::id_type id_type;

struct source_location
{ unsigned int start_line;
  unsigned short extent, first_col, last_col;
  id_type file;
  source_location(const YYLTYPE& loc); // construct from parser-provided data
  source_location() : start_line(~0u) {}
    // sometimes we have no location
  source_location
    (const source_location& left, const source_location& right);
    // compute span
  bool undefined() const { return start_line==~0u; }
};

typedef containers::simple_list<expr> expr_list;
typedef containers::sl_node<expr>* raw_expr_list;
  // raw counterpart for use by parser

typedef containers::weak_sl_list_const_iterator<expr> wel_const_iterator;
typedef containers::weak_sl_list_iterator<expr> wel_iterator;
  

typedef struct application_node* app;

typedef containers::simple_list<struct id_pat> patlist;
typedef containers::sl_node<struct id_pat>* raw_patlist;

struct raw_id_pat
{ id_type name;
  unsigned char kind;
  // bits 0,1,2: whether pattern has a name, has sublist, is const
  raw_patlist sublist;
};
struct id_pat
{ id_type name;
  unsigned char kind;
  // bits 0,1,2: whether pattern has a name, has sublist, is const
  patlist sublist;

  id_pat() :name(), kind(0x0), sublist() {}
  explicit id_pat(const raw_id_pat& x)
   : name(x.name), kind(x.kind)
  , sublist((kind & 0x2)==0 ? nullptr : x.sublist)
  {}
  id_pat (id_type n) : name(n), kind(0x1), sublist() {}
    // name only,  no constness
  id_pat(patlist&& l): name(), kind(0x2),  sublist(std::move(l)) {}
    // no name, no constness
  id_pat (id_type n, unsigned char k, patlist&& l)
  : name(n), kind(k), sublist(std::move(l)) {}

  id_pat (const id_pat& x) =  delete ;
id_pat& operator=(const id_pat& x) =  delete ;
#ifdef incompletecpp11
  id_pat (id_pat&& x)
  : name(x.name), kind(x.kind), sublist(std::move(x.sublist)) {}
id_pat& operator=(id_pat&& x)
  { name = x.name; kind = x.kind; sublist = std::move(x.sublist);
    return *this;
  }
#else
id_pat (id_pat&& x) =  default ;
id_pat& operator=(id_pat&& x) =  default ;
#endif
  raw_id_pat release()
  { return { name, kind, sublist.release() }; }
};

typedef struct let_expr_node* let;

typedef struct lambda_node* lambda;

typedef struct conditional_node* cond;

typedef struct while_node* w_loop;
typedef struct for_node* f_loop;
typedef struct cfor_node* c_loop;

typedef struct subscription_node* sub;
typedef struct slice_node* slc;

typedef struct cast_node* cast;

typedef struct op_cast_node* op_cast;

typedef struct assignment_node* assignment;

typedef struct comp_assignment_node* comp_assignment;
typedef struct field_assignment_node* fld_assignment;

typedef struct sequence_node* sequence;


enum expr_kind 
 { 
   integer_denotation, string_denotation, boolean_denotation, 
   
   applied_identifier,
   last_value_computed,break_expr,return_expr,
   die_expr, 
   
    tuple_display, list_display, 
   
    function_call, 
   
   negation_expr, 
   
    let_expr, 
   
   lambda_expr,
   
    conditional_expr, 
   
    int_case_expr, 
   
    while_expr, for_expr, cfor_expr, 
   
    subscription, slice, 
   
   cast_expr, 
   
   op_cast_expr, 
   
   ass_stat, 
   
   comp_ass_stat, field_ass_stat, 
   
   seq_expr, next_expr, do_expr, 
 no_expr };
typedef struct expr* expr_p; // raw pointer type for use on parser stack
struct expr {
  expr_kind kind;
  union { 
          int int_denotation_variant;
          bool bool_denotation_variant;
          #ifdef incompletecpp11
          const char* str_denotation_variant;
          #else
          std::string str_denotation_variant;
          #endif
          
          id_type identifier_variant;
          unsigned break_variant;
          expr_p return_variant;
          
          raw_expr_list sublist;
          
          app call_variant;
          
          expr* negation_variant;
          
          let let_variant;
          
          lambda lambda_variant;
          
          cond if_variant;
          
          w_loop while_variant;
          f_loop for_variant;
          c_loop cfor_variant;
          
          sub subscription_variant;
          slc slice_variant;
          
          cast cast_variant;
          
          op_cast op_cast_variant;
          
          assignment assign_variant;
          
          comp_assignment comp_assign_variant;
          fld_assignment field_assign_variant;
          
          sequence sequence_variant;
 };
  source_location loc;

  
  expr() : kind(no_expr), loc() {}
  expr(const expr& x) = delete;
  expr& operator=(const expr& x) = delete;
  ~expr(); 
  
  void set_from (expr& other);
  expr (expr&& other);
  void operator= (expr&& other);
  void swap (expr& other);
  
    struct int_tag {}; 
    struct bool_tag {};
    expr(int n, const YYLTYPE& loc, int_tag)
  : kind(integer_denotation), int_denotation_variant(n), loc(loc) {}
    expr (bool b, const YYLTYPE& loc, bool_tag)
  : kind(boolean_denotation), bool_denotation_variant(b), loc(loc) {}
    expr(std::string&& s, const YYLTYPE& loc)
     : kind(string_denotation)
  #ifdef incompletecpp11
     , str_denotation_variant(std::strcpy(new char[s.size()+1],s.c_str()))
  #else
     , str_denotation_variant(std::move(s))
  #endif
     , loc(loc) {}
  
    struct identifier_tag {}; 
    struct dollar_tag {};
    struct break_tag { unsigned depth; };
    struct return_tag {};
    struct die_tag {};
  
    expr(id_type id, const YYLTYPE& loc, identifier_tag)
  : kind(applied_identifier), identifier_variant(id), loc(loc) {}
    expr (const YYLTYPE& loc, dollar_tag)
    : kind(last_value_computed), loc(loc) {}
    expr (const YYLTYPE& loc, break_tag t)
    : kind(break_expr), break_variant(t.depth), loc(loc) {}
    expr (expr_p exp, const YYLTYPE& loc, return_tag t)
    : kind(return_expr), return_variant(exp), loc(loc) {}
    expr (const YYLTYPE& loc, die_tag)
    : kind(die_expr), loc(loc) {}
  
  struct tuple_display_tag  {}; struct list_display_tag {};
  expr(expr_list&& nodes, tuple_display_tag, const YYLTYPE& loc)
  : kind(tuple_display)
    , sublist(nodes.release())
    , loc(loc)
  {}
  expr(expr_list&& nodes, list_display_tag, const YYLTYPE& loc)
  : kind(list_display)
    , sublist(nodes.release())
    , loc(loc)
  {}
  expr(expr_list&& nodes, tuple_display_tag, const source_location& loc)
  : kind(tuple_display)
    , sublist(nodes.release())
    , loc(loc)
  {}
  expr(expr_list&& nodes, tuple_display_tag);
  
  expr(app&& fx, const YYLTYPE& loc)
  : kind(function_call)
   , call_variant(std::move(fx))
   , loc(loc)
   {}
  expr(app&& fx, const source_location& loc)
  : kind(function_call)
   , call_variant(std::move(fx))
   , loc(loc)
   {}
  
  struct negate_tag {}; 
  expr(expr* p, const YYLTYPE& loc,negate_tag)
   : kind(negation_expr)
   , negation_variant(p)
   , loc(loc)
  {}
  
  expr(let&& declaration, const YYLTYPE& loc)
   : kind(let_expr)
   , let_variant(std::move(declaration))
   , loc(loc)
   {}
  
  expr(lambda&& fun, const YYLTYPE& loc)
   : kind(lambda_expr)
   , lambda_variant(std::move(fun))
   , loc(loc)
  {}
  
  expr(expr_kind which, cond&& conditional, const YYLTYPE& loc)
   : kind(which)
   , if_variant(std::move(conditional))
   , loc(loc)
  {}
  
  expr(w_loop&& loop, const YYLTYPE& loc)
   : kind(while_expr)
   , while_variant(std::move(loop))
   , loc(loc)
  {}
  expr(f_loop&& loop, const YYLTYPE& loc)
   : kind(for_expr)
   , for_variant(std::move(loop))
   , loc(loc)
  {}
  expr(c_loop&& loop, const YYLTYPE& loc)
   : kind(cfor_expr)
   , cfor_variant(std::move(loop))
   , loc(loc)
  {}
  
  expr(sub&& s, const YYLTYPE& loc)
   : kind(subscription)
   , subscription_variant(std::move(s))
   , loc(loc)
   {}
  expr(sub&& s, const source_location& loc)
   : kind(subscription)
   , subscription_variant(std::move(s))
   , loc(loc)
   {}
  expr(slc&& s, const YYLTYPE& loc)
   : kind(slice)
   , slice_variant(std::move(s))
   , loc(loc)
   {}
  
  expr(cast&& c, const YYLTYPE& loc)
   : kind(cast_expr)
   , cast_variant(std::move(c))
   , loc(loc)
  {}
  
  expr(op_cast&& c, const YYLTYPE& loc)
   : kind(op_cast_expr)
   , op_cast_variant(std::move(c))
   , loc(loc)
  {}
  
  expr(assignment&& a, const YYLTYPE& loc)
   : kind(ass_stat)
   , assign_variant(std::move(a))
   , loc(loc)
  {}
  
  expr(comp_assignment ca, const YYLTYPE& loc)
   : kind(comp_ass_stat)
   , comp_assign_variant(ca)
   , loc(loc)
  {}
  
  expr(fld_assignment&& fa, const YYLTYPE& loc)
   : kind(field_ass_stat)
   , field_assign_variant(std::move(fa))
   , loc(loc)
  {}
  
  expr(sequence&& s, unsigned which, const YYLTYPE& loc)
   : kind(which==0 ? seq_expr : which==1 ? next_expr : do_expr)
   , sequence_variant(std::move(s))
   , loc(loc)
  {}

};
typedef std::unique_ptr<expr> expr_ptr;


struct application_node
{ expr fun;  expr arg;

  application_node(expr&& fun, expr&& arg)
  : fun(std::move(fun)), arg(std::move(arg)) {}
};

struct formula_node { expr left_subtree; expr op_exp; int prio;
formula_node(expr&& l,expr&& o, int prio)
 : left_subtree(std::move(l)), op_exp(std::move(o)), prio(prio) {}
#ifdef incompletecpp11
formula_node(const formula_node& x) = delete;
formula_node operator=(const formula_node& x) = delete;
#endif
};

typedef containers::mirrored_simple_list<formula_node> form_stack;
typedef containers::sl_node<formula_node>* raw_form_stack;

struct let_pair { id_pat pattern; expr val;
#ifdef incompletecpp11
  let_pair (const let_pair&) = delete;
  let_pair& operator=(const let_pair&) = delete;
  let_pair (let_pair&& x)
  : pattern(std::move(x.pattern)), val(std::move(x.val)) {}
  let_pair (id_pat&& p, expr&& v)
  : pattern(std::move(p)), val(std::move(v)) {}
#endif
};
typedef containers::simple_list<let_pair> let_list;
typedef containers::sl_node<let_pair>* raw_let_list;

struct let_expr_node
{ id_pat pattern; expr val; expr body;

  let_expr_node(id_pat&& pattern, expr&& val, expr&& body)
: pattern(std::move(pattern)), val(std::move(val)), body(std::move(body)){}
};

struct lambda_node
{ id_pat pattern; type_expr parameter_type; expr body;

  lambda_node(id_pat&& pattern, type_expr&& type, expr&& body)
  : pattern(std::move(pattern))
  , parameter_type(std::move(type))
  , body(std::move(body))
{}
};

struct conditional_node
{ expr condition; containers::sl_node<expr> branches;

  conditional_node(expr&& condition, expr&& then_branch, expr&& else_branch)
: condition(std::move(condition))
  , branches(std::move(then_branch))
  { branches.next.reset(new
      containers::sl_node<expr>(std::move(else_branch)));
  }
  
  conditional_node(expr&& condition, containers::sl_node<expr>&& branches)
  : condition(std::move(condition))
    , branches(std::move(branches))
    {}

};

struct while_node
{ expr body; // contains a |do|-expression producing condition and body
  BitSet<2> flags;

  while_node(expr&& body, unsigned flags)
: body(std::move(body))
  , flags(flags){}
  // backward compatibility for gcc 4.6
};
struct for_node
{ id_pat id;  expr in_part;  expr body;
  BitSet<2> flags;

  for_node(id_pat&& id, expr&& in_part, expr&& body, unsigned flags)
: id(std::move(id))
  , in_part(std::move(in_part))
  , body(std::move(body))
  , flags(flags){}
};
struct cfor_node
{ id_type id;  expr count;  expr bound;  expr body;
  BitSet<3> flags;

  cfor_node
     (id_type id, expr&& count, expr&& bound, expr&& body, unsigned flags)
: id(id)
  , count(std::move(count))
  , bound(std::move(bound))
  , body(std::move(body))
  , flags(flags){}
};

struct subscription_node
{ expr array; expr index; bool reversed;

  subscription_node(expr&& array, expr&& index, bool reversed)
: array(std::move(array))
  , index(std::move(index))
  , reversed(reversed){}
};
struct slice_node
{ expr array; expr lower,upper; BitSet<3> flags;

  slice_node(expr&& array, expr&& lower, expr&& upper, unsigned flags)
: array(std::move(array))
  , lower(std::move(lower))
  , upper(std::move(upper))
  , flags(flags)
  {}
};

struct cast_node
{ type_expr type; expr exp;

  cast_node(type_expr&& type, expr&& exp)
: type(std::move(type))
  , exp(std::move(exp)){}
  // backward compatibility for gcc 4.6
};

struct op_cast_node
{ id_type oper; type_expr type;

  op_cast_node(id_type oper,type_expr&& type)
: oper(oper)
  , type(std::move(type)){}
  // backward compatibility for gcc 4.6
};

struct assignment_node
{ id_pat lhs; expr rhs;

  assignment_node(id_type lhs, expr&& rhs)
: lhs(lhs)
  , rhs(std::move(rhs)){}
  assignment_node(id_pat&& lhs, expr&& rhs)
: lhs(std::move(lhs))
  , rhs(std::move(rhs)){}
#ifdef incompletecpp11
  assignment_node(const assignment_node& x) = delete;
  assignment_node(assignment_node&& x)
: lhs(std::move(x.lhs))
  , rhs(std::move(x.rhs))
  {}
#endif
};

struct comp_assignment_node
{ id_type aggr; expr index; expr rhs; bool reversed;

  comp_assignment_node(id_type aggr, expr&& index, expr&& rhs, bool reversed)
: aggr(aggr)
  , index(std::move(index))
  , rhs(std::move(rhs))
  , reversed(reversed)
  {}
#ifdef incompletecpp11
  comp_assignment_node(const comp_assignment_node& x) = delete;
  comp_assignment_node(comp_assignment_node&& x)
: aggr(x.aggr)
  , index(std::move(x.index))
  , rhs(std::move(x.rhs))
  , reversed(x.reversed)
  {}
#endif
};

struct field_assignment_node
{ id_type aggr, selector; expr rhs;

  field_assignment_node(id_type aggr, id_type selector, expr&& rhs)
: aggr(aggr)
  , selector(selector)
  , rhs(std::move(rhs))
  {}
#ifdef incompletecpp11
  field_assignment_node(const field_assignment_node& x) = delete;
  field_assignment_node(field_assignment_node&& x)
: aggr(x.aggr)
  , selector(selector)
  , rhs(std::move(x.rhs))
  {}
#endif
};

struct sequence_node
{ expr first; expr last;

  sequence_node(expr&& first, expr&& last)
: first(std::move(first))
  , last(std::move(last)){}
  // backward compatibility for gcc 4.6
};


  }
}
#endif

